/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __draw_DrawMode_h__
#define __draw_DrawMode_h__

namespace fe
{

/**************************************************************************//**
	@brief Configuration for rendering such as line width and backface culling

	@ingroup draw

	An instance of a DrawMode is shared between all draw operations that
	occur while it is the current DrawMode to a DrawI.
	Changing a DrawMode at any time can affect all elements that are
	associated with that DrawMode.

	It is intended for these modes to be configured and left alone prior
	to being used.  They are not an indirect interface extension to DrawI.

	The DrawMode is an optimization for cached rendering.  Without them,
	either every draw operation would have to cache an entire drawing
	state or there would be an extraordinary amount of cross checking
	to match up coincidentally identical states.
*//***************************************************************************/
class FE_DL_EXPORT DrawMode: public Counted
{
	public:

					/** @brief Rendering schemes that an an
							implementation @em may recognize

						Not all methods may be available in a
						particular situation.  */
		enum		DrawStyle
					{
						e_defaultStyle,
						e_pointCloud,
						e_wireframe,
						e_outline,
						e_solid,
						e_edgedSolid,
						e_foreshadow,
						e_count
					};

					DrawMode(void):
						m_drawStyle(e_defaultStyle),
						m_layer(0),
						m_pointSize(1.0f),
						m_lineWidth(1.0f),
						m_antialias(FALSE),
						m_frontfaceCulling(FALSE),
						m_backfaceCulling(TRUE),
						m_twoSidedLighting(FALSE),
						m_zBuffering(TRUE),
						m_lit(TRUE)
					{	setName("DrawMode"); }

		void		copy(sp<DrawMode> a_spDrawMode)
					{
						m_spTextureImage=a_spDrawMode->m_spTextureImage;
						m_drawStyle=a_spDrawMode->m_drawStyle;
						m_group=a_spDrawMode->m_group;
						m_layer=a_spDrawMode->m_layer;
						m_pointSize=a_spDrawMode->m_pointSize;
						m_lineWidth=a_spDrawMode->m_lineWidth;
						m_antialias=a_spDrawMode->m_antialias;
						m_frontfaceCulling=a_spDrawMode->m_frontfaceCulling;
						m_backfaceCulling=a_spDrawMode->m_backfaceCulling;
						m_twoSidedLighting=a_spDrawMode->m_twoSidedLighting;
						m_zBuffering=a_spDrawMode->m_zBuffering;
						m_lit=a_spDrawMode->m_lit;
					}

					/// @brief Set raster for texture mapping
		void		setTextureImage(sp<ImageI> spImageI)
					{	m_spTextureImage=spImageI; }

					/// @brief Get raster for texture mapping
		sp<ImageI>	textureImage(void) const
					{	return m_spTextureImage; }

					/// @brief Set rendering scheme (see DrawStyle)
		void		setDrawStyle(DrawStyle style)	{ m_drawStyle=style; }

					/// @brief Get rendering scheme (see DrawStyle)
		DrawStyle	drawStyle(void) const			{ return m_drawStyle; }

					/// @brief Set named group
		void		setGroup(String a_group)		{ m_group=a_group; }

					/// @brief Get named group
		String		group(void) const				{ return m_group; }

					/// @brief Set rendering ordering (highest last)
		void		setLayer(I32 set)				{ m_layer=set; }

					/// @brief Get rendering ordering (highest last)
		I32			layer(void) const				{ return m_layer; }

					/// @brief Set diameter of all following drawn points
		void		setPointSize(Real pixels)
					{	m_pointSize=pixels; }

					/// @brief Get diameter of all following drawn points
		Real		pointSize(void) const
					{	return m_pointSize; }

					/// @brief Set width of all following drawn lines
		void		setLineWidth(Real pixels)
					{	m_lineWidth=pixels; }

					/// @brief Get width of all following drawn lines
		Real		lineWidth(void) const
					{	return m_lineWidth; }

					/// @brief Set rasterization anti-aliasing
		void		setAntialias(BWORD antialias)
					{	m_antialias=antialias; }

					/// @brief Get whether anti-aliasing is used
		BWORD		antialias(void) const
					{	return m_antialias; }

					/// @brief Set whether frontfaces are removed
		void		setFrontfaceCulling(BWORD frontfaceCulling)
					{	m_frontfaceCulling=frontfaceCulling; }

					/// @brief Get whether frontfaces are removed
		BWORD		frontfaceCulling(void) const
					{	return m_frontfaceCulling; }

					/// @brief Set whether backfaces are removed
		void		setBackfaceCulling(BWORD backfaceCulling)
					{	m_backfaceCulling=backfaceCulling; }

					/// @brief Get whether backfaces are removed
		BWORD		backfaceCulling(void) const
					{	return m_backfaceCulling; }

					/** @brief Set whether backfaces have an independent
						material */
		void		setTwoSidedLighting(BWORD twoSidedLighting)
					{	m_twoSidedLighting=twoSidedLighting; }

					/** @brief Get whether backfaces have an independent
						material */
		BWORD		twoSidedLighting(void) const
					{	return m_twoSidedLighting; }

					/// @brief Set whether z-buffering is used
		void		setZBuffering(BWORD zBuffering)
					{	m_zBuffering=zBuffering; }

					/// @brief Get whether z-buffering is used
		BWORD		zBuffering(void) const
					{	return m_zBuffering; }

					/** @brief Set lit */
		void		setLit(BWORD lit)
					{	m_lit=lit; }

					/// @brief Get whether lit
		BWORD		lit(void) const
					{	return m_lit; }

static	DrawStyle	stringToStyle(const String &a_styleName);
	private:

		sp<ImageI>		m_spTextureImage;
		DrawStyle		m_drawStyle;
		String			m_group;
		I32				m_layer;
		Real			m_pointSize;
		Real			m_lineWidth;
		BWORD			m_antialias;
		BWORD			m_frontfaceCulling;
		BWORD			m_backfaceCulling;
		BWORD			m_twoSidedLighting;
		BWORD			m_zBuffering;
		BWORD			m_lit;
};

static char DrawStyleText[][32] =
{
	"default",
	"pointCloud",
	"wireframe",
	"outline",
	"solid",
	"edgedSolid",
	"foreshadow"
};

inline String print(const sp<DrawMode> spDrawMode)
{
	String s;
	s.sPrintf("style \"%s\" layer %d pointSize %g lineWidth %g"
			" aa %d backface %d twoSided %d z %d lit %d",
			DrawStyleText[spDrawMode->drawStyle()%DrawMode::e_count],
			spDrawMode->layer(),
			spDrawMode->pointSize(),
			spDrawMode->lineWidth(),
			spDrawMode->pointSize(),
			spDrawMode->antialias(),
			spDrawMode->backfaceCulling(),
			spDrawMode->twoSidedLighting(),
			spDrawMode->zBuffering(),
			spDrawMode->lit());
	return s;
}

inline DrawMode::DrawStyle DrawMode::stringToStyle(const String &a_styleName)
{
	if(a_styleName == "pointCloud") { return e_pointCloud; }
	if(a_styleName == "wireframe") { return e_wireframe; }
	if(a_styleName == "outline") { return e_outline; }
	if(a_styleName == "edgedSolid") { return e_edgedSolid; }
	if(a_styleName == "solid") { return e_solid; }
	if(a_styleName == "foreshadow") { return e_foreshadow; }

	return e_defaultStyle;
}
} // namespace

#endif /* __draw_DrawMode_h__ */
