/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "lua.pmh"

namespace fe
{

void *lua_alloc (void *ud, void *ptr, size_t osize, size_t nsize)
{
  (void)ud;
  (void)osize;
  if (nsize == 0) {
    free(ptr);
    return NULL;
  }
  else
    return realloc(ptr, nsize);
}

LuaContext::LuaContext(sp<Master> spMaster, lua_State *pLuaState):
		m_name("LuaContext")
{
	if(!spMaster.isValid())
	{
		feX("LuaContext::LuaContext",
			"Master not valid");
	}

	registerRegion(this,sizeof(*this));

	m_hpMaster = spMaster;
	sp<TypeMaster> spTM = spMaster->typeMaster();
	m_pLuaState = pLuaState;
	m_spIntegerType = spTM->lookupType<int>();
	m_spFloatType = spTM->lookupType<float>();
	m_spDoubleType = spTM->lookupType<double>();
	m_spStringType = spTM->lookupType<String>();
	m_spRecordType = spTM->lookupType<Record>();
	m_spGroupType = spTM->lookupType<sp<RecordGroup> >();
	m_spArrayType = spTM->lookupType<sp<RecordArray> >();
	m_spComponentType = spTM->lookupType<sp<Component> >();
	m_spIntArrayType = spTM->lookupType< Array<I32> >();
	m_spVector2Type = spTM->lookupType<Vector2f>();
	m_spVector3Type = spTM->lookupType<Vector3>();
	m_spVector4Type = spTM->lookupType<Vector4>();
	m_spVector4fType = spTM->lookupType<Vector4f>();
	m_spLayoutType = spTM->lookupType<sp<Layout> >();

	m_luaTypes.resize(e_size);

	m_luaTypes[e_record] =
		new LuaRecordType(this,"mt_record");
	m_luaTypes[e_record]->putMetaTable();

	m_luaTypes[e_layout] =
		new LuaLayoutType(this,"mt_layout");
	m_luaTypes[e_layout]->putMetaTable();

	m_luaTypes[e_scope] =
		new LuaScopeType(this,"mt_scope");
	m_luaTypes[e_scope]->putMetaTable();

	m_luaTypes[e_array] =
		new LuaArrayType(this,"mt_array");
	m_luaTypes[e_array]->putMetaTable();

	m_luaTypes[e_arrayit] =
		new LuaArrayItType(this,"mt_arrayit");
	m_luaTypes[e_arrayit]->putMetaTable();

	m_luaTypes[e_group] =
		new LuaGroupType(this,"mt_group");
	m_luaTypes[e_group]->putMetaTable();

	m_luaTypes[e_groupit] =
		new LuaGroupItType(this,"mt_groupit");
	m_luaTypes[e_groupit]->putMetaTable();

	m_luaTypes[e_component] =
		new LuaComponentType(this,"mt_component");
	m_luaTypes[e_component]->putMetaTable();

	m_luaTypes[e_intarray] =
		new LuaIntArrayType(this,"mt_intarray");
	m_luaTypes[e_intarray]->putMetaTable();

	m_luaTypes[e_intarrayit] =
		new LuaIntArrayItType(this,"mt_intarrayit");
	m_luaTypes[e_intarrayit]->putMetaTable();

	m_luaTypes[e_vector] =
		new LuaVectorType(this,"mt_vector");
	m_luaTypes[e_vector]->putMetaTable();

	m_luaTypes[e_vector4] =
		new LuaVector4Type(this,"mt_vector4");
	m_luaTypes[e_vector4]->putMetaTable();

	m_luaTypes[e_vector4f] =
		new LuaVector4fType(this,"mt_vector4f");
	m_luaTypes[e_vector4f]->putMetaTable();

	m_luaTypes[e_vector3] =
		new LuaVector3Type(this,"mt_vector3");
	m_luaTypes[e_vector3]->putMetaTable();

	m_luaTypes[e_vector2] =
		new LuaVector2Type(this,"mt_vector2");
	m_luaTypes[e_vector2]->putMetaTable();

	lua_pushstring(m_pLuaState, "fe_master");
	lua_pushlightuserdata(m_pLuaState, spMaster.raw());
	lua_settable(m_pLuaState, LUA_GLOBALSINDEX);

	lua_pushstring(m_pLuaState, "fe_context");
	lua_pushlightuserdata(m_pLuaState, this);
	lua_settable(m_pLuaState, LUA_GLOBALSINDEX);

	setWindowEventEnums();

	lua_register(m_pLuaState, "manage",				manage);
	lua_register(m_pLuaState, "catalog",			catalog);
	lua_register(m_pLuaState, "create",				create);
	lua_register(m_pLuaState, "createRecordArray",	createRecordArray);
	lua_register(m_pLuaState, "createScope",		createScope);
	lua_register(m_pLuaState, "componentScope",		createScopeComponent);
	lua_register(m_pLuaState, "b_or",				b_or);
	lua_register(m_pLuaState, "b_and",				b_and);
	lua_register(m_pLuaState, "createRecordGroup",	createRecordGroup);
	lua_register(m_pLuaState, "createRecord",		createRecord);
}

LuaContext::~LuaContext(void)
{
//	feLog("~LuaContext\n");

#if FE_COUNTED_TRACK
	U32 size=m_trackArray.size();
	for(U32 m=0;m<size;m++)
	{
		deregisterRegion(m_trackArray[m]);
	}
#endif
}

void LuaContext::set(const char *tbl, const char *name, double value)
{
	lua_pushstring(m_pLuaState, tbl);
	lua_gettable(m_pLuaState, LUA_GLOBALSINDEX);
	if(!lua_istable(m_pLuaState, -1))
	{
		lua_pop(m_pLuaState, 1);
		lua_pushstring(m_pLuaState, tbl);
		lua_newtable(m_pLuaState);
		lua_settable(m_pLuaState, LUA_GLOBALSINDEX);
		lua_pushstring(m_pLuaState, tbl);
		lua_gettable(m_pLuaState, LUA_GLOBALSINDEX);
	}
	lua_pushstring(m_pLuaState, name);
	lua_pushnumber(m_pLuaState, value);
	lua_settable(m_pLuaState, -3);
	lua_pop(m_pLuaState, 1);
}


sp<LuaBaseType> LuaContext::lookup(FE_UWORD tid)
{
	return m_luaTypes[tid];
}

lua_State *LuaContext::state(void)
{
	return m_pLuaState;
}


void LuaContext::pushInstance(Instance &instance)
{
	if(instance.type() == m_spIntegerType)
	{
		lua_pushnumber(m_pLuaState, (lua_Number)(instance.cast<int>()));
	}
	else if(instance.type() == m_spFloatType)
	{
		lua_pushnumber(m_pLuaState, (lua_Number)(instance.cast<float>()));
	}
	else if(instance.type() == m_spDoubleType)
	{
		lua_pushnumber(m_pLuaState, (lua_Number)(instance.cast<double>()));
	}
	else if(instance.type() == m_spStringType)
	{
		lua_pushstring(m_pLuaState, instance.cast<String>().c_str());
	}
	else if(instance.type() == m_spRecordType)
	{
		t_lua_record::push(this, instance.cast<Record>());

// TODO restore for RecordAV
#if FALSE //FE_COUNTED_TRACK
		Record& record=instance.cast<Record>();
		if(record.isValid())
		{
			void* data=record.data();
			U32 size=record.layout()->size();

			Counted::registerRegion(data,size,
					"LuaContext:Record "+record.layout()->name());
			Counted::trackReference(data,dynamic_cast<Counted*>(this),
					"LuaContext");

			m_trackArray.push_back(data);
		}
#endif
	}
	else if(instance.type() == m_spGroupType)
	{
		t_lua_group::push(this, instance.cast<sp<RecordGroup> >());
	}
	else if(instance.type() == m_spArrayType)
	{
		t_lua_array::push(this, instance.cast<sp<RecordArray> >());
	}
	else if(instance.type() == m_spComponentType)
	{
		LuaComponentObject::pushComponent(state(),
			instance.cast<sp<Component> >());
	}
	else if(instance.type() == m_spLayoutType)
	{
		t_lua_layout::push(this, instance.cast<sp<Layout> >());
	}
	else if(instance.type() == m_spIntArrayType)
	{
		t_lua_intarray::push(this, &instance.cast< Array<I32> >());
	}
	else if(instance.type() == m_spVector2Type)
	{
		t_lua_vector::push(this, instance.cast<Vector2f>().raw());
	}
	else if(instance.type() == m_spVector3Type)
	{
		t_lua_vector3::push(this, &instance.cast<Vector3>());
	}
	else if(instance.type() == m_spVector4Type)
	{
		t_lua_vector4::push(this,&instance.cast<Vector4>());
	}
	else if(instance.type() == m_spVector4fType)
	{
		t_lua_vector4f::push(this,&instance.cast<Vector4f>());
	}
	else
	{
		sp<TypeMaster> spTM = m_hpMaster->typeMaster();
		String errstr;
		errstr.sPrintf("attempt to push unsupported type %s",
			spTM->reverseLookup(instance.type()).c_str());
		error(m_pLuaState, errstr.c_str());
	}
}

Instance LuaContext::popCreateInstance(void)
{
	Instance instance;
	int t = lua_type(m_pLuaState, -1);
	if(t == LUA_TNUMBER)
	{
		instance.create<Real>(typeMaster());
		instance.cast<Real>() = (Real)lua_tonumber(m_pLuaState, -1);
		lua_pop(m_pLuaState, 1);
	}
	else if(t == LUA_TSTRING)
	{
		instance.create<String>(typeMaster());
		instance.cast<String>() = (String)lua_tostring(m_pLuaState, -1);
		lua_pop(m_pLuaState, 1);
	}
	else
	{
	}

	return instance;
}

void LuaContext::popInstance(Instance &instance, bool create)
{
	int t = lua_type(m_pLuaState, -1);
	bool mismatch = false;

	if(!instance.type().isValid() && create)
	{
		if(t == LUA_TNUMBER)
		{
			instance.create<double>(typeMaster());
		}

		if(t == LUA_TSTRING)
		{
			instance.create<String>(typeMaster());
		}

	}

	if(instance.type() == m_spIntegerType)
	{
		if(t == LUA_TNUMBER || t == LUA_TSTRING)
		{
			instance.cast<int>() = (int)lua_tonumber(m_pLuaState, -1);
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spFloatType)
	{
		if(t == LUA_TNUMBER || t == LUA_TSTRING)
		{
			instance.cast<float>() = (float)lua_tonumber(m_pLuaState, -1);
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spDoubleType)
	{
		if(t == LUA_TNUMBER || t == LUA_TSTRING)
		{
			instance.cast<double>() = (double)lua_tonumber(m_pLuaState, -1);
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spStringType)
	{
		if(t == LUA_TNUMBER || t == LUA_TSTRING)
		{
			instance.cast<String>() = lua_tostring(m_pLuaState, -1);
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spRecordType)
	{
		if(t == LUA_TUSERDATA)
		{
			void *pV = lua_touserdata(m_pLuaState, -1);
			if(((LuaBaseObject *)pV)->type() == m_luaTypes[e_record])
			{
				t_lua_record *pLR = (t_lua_record *)pV;
				instance.cast<Record>() = pLR->data();

// TODO restore for RecordAV
#if FALSE //FE_COUNTED_TRACK
				Record& record=instance.cast<Record>();
				if(record.isValid())
				{
					feLog("LuaContext dereg\n");
					Counted::deregisterRegion(record.data());
				}
#endif
			}
			else { mismatch = true; }
		}
		else if(t == LUA_TNIL)
		{
			Record empty;
			instance.cast<Record>() = empty;
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spGroupType)
	{
		if(t == LUA_TUSERDATA)
		{
			void *pV = lua_touserdata(m_pLuaState, -1);
			if(((LuaBaseObject *)pV)->type() == m_luaTypes[e_group])
			{
				t_lua_group *pLR = (t_lua_group *)pV;
				instance.cast<sp<RecordGroup> >() = pLR->data();
			}
			else { mismatch = true; }
		}
		else if(t == LUA_TNIL)
		{
			instance.cast<sp<RecordGroup> >() = NULL;
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spArrayType)
	{
		if(t == LUA_TUSERDATA)
		{
			void *pV = lua_touserdata(m_pLuaState, -1);
			if(((LuaBaseObject *)pV)->type() == m_luaTypes[e_array])
			{
				t_lua_array *pLR = (t_lua_array *)pV;
				instance.cast<sp<RecordArray> >() = pLR->data();
			}
			else { mismatch = true; }
		}
		else if(t == LUA_TNIL)
		{
			instance.cast<sp<RecordArray> >() = NULL;
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spComponentType)
	{
		if(t == LUA_TUSERDATA)
		{
			void *pV = lua_touserdata(m_pLuaState, -1);
			if(((LuaBaseObject *)pV)->type() == m_luaTypes[e_component])
			{
				t_lua_component *pLR = (t_lua_component *)pV;
				instance.cast<sp<Component> >() = pLR->data();
			}
			else { mismatch = true; }
		}
		else if(t == LUA_TNIL)
		{
			instance.cast<sp<Component> >() = NULL;
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spLayoutType)
	{
		if(t == LUA_TUSERDATA)
		{
			void *pV = lua_touserdata(m_pLuaState, -1);
			if(((LuaBaseObject *)pV)->type() == m_luaTypes[e_layout])
			{
				t_lua_layout *pLR = (t_lua_layout *)pV;
				instance.cast<sp<Layout> >() = pLR->data();
			}
			else { mismatch = true; }
		}
		else if(t == LUA_TNIL)
		{
			instance.cast<sp<Layout> >() = NULL;
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spIntArrayType)
	{
		if(t == LUA_TUSERDATA)
		{
			void *pV = lua_touserdata(m_pLuaState, -1);
			if(((LuaBaseObject *)pV)->type() == m_luaTypes[e_intarray])
			{
				t_lua_intarray *pLR = (t_lua_intarray *)pV;
				instance.cast< Array<I32> >() = *pLR->data();
			}
			else { mismatch = true; }
		}
		else if(t == LUA_TNIL)
		{
			instance.cast< Array<I32> >().clear();
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spVector4Type)
	{
		if(t == LUA_TUSERDATA)
		{
			void *pV = lua_touserdata(m_pLuaState, -1);
			if(((LuaBaseObject *)pV)->type() == m_luaTypes[e_vector4])
			{
				t_lua_vector4 *pLV = (t_lua_vector4 *)pV;
				instance.cast<Vector4>() = *pLV->data();
			}
			else { mismatch = true; }
		}
		else if(t == LUA_TNIL)
		{
			instance.cast<Vector4>() = NULL;
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spVector4fType)
	{
		if(t == LUA_TUSERDATA)
		{
			void *pV = lua_touserdata(m_pLuaState, -1);
			if(((LuaBaseObject *)pV)->type() == m_luaTypes[e_vector4f])
			{
				t_lua_vector4f *pLV = (t_lua_vector4f *)pV;
				instance.cast<Vector4f>() = *pLV->data();
			}
			else { mismatch = true; }
		}
		else if(t == LUA_TNIL)
		{
			instance.cast<Vector4f>() = NULL;
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spVector3Type)
	{
		if(t == LUA_TUSERDATA)
		{
			void *pV = lua_touserdata(m_pLuaState, -1);
			if(((LuaBaseObject *)pV)->type() == m_luaTypes[e_vector3])
			{
				t_lua_vector3 *pLV = (t_lua_vector3 *)pV;
				instance.cast<Vector3>() = *pLV->data();
			}
			else { mismatch = true; }
		}
		else if(t == LUA_TNIL)
		{
			instance.cast<Vector3>() = NULL;
		}
		else { mismatch = true; }
	}
	else if(instance.type() == m_spVector2Type)
	{
		if(t == LUA_TUSERDATA)
		{
			void *pV = lua_touserdata(m_pLuaState, -1);
			if(((LuaBaseObject *)pV)->type() == m_luaTypes[e_vector2])
			{
				t_lua_vector2 *pLV = (t_lua_vector2 *)pV;
				instance.cast<Vector2>() = *pLV->data();
			}
			else { mismatch = true; }
		}
		else if(t == LUA_TNIL)
		{
			instance.cast<Vector2>() = NULL;
		}
		else { mismatch = true; }
	}
	else
	{
		sp<TypeMaster> spTM = m_hpMaster->typeMaster();
		String errstr;
		errstr.sPrintf("lua->fe attempt to pop unsupported type %s",
			spTM->reverseLookup(instance.type()).c_str());
		error(m_pLuaState, errstr.c_str());
	}

	if(mismatch)
	{
		sp<TypeMaster> spTM = m_hpMaster->typeMaster();
		String errstr;
		errstr.sPrintf("lua->fe type mismatch \'%s\'",
			spTM->reverseLookup(instance.type()).c_str());
		error(m_pLuaState, errstr.c_str());
	}

	lua_pop(m_pLuaState, 1);
}

int LuaContext::createRecordGroup(lua_State *pLuaState)
{
	lua_pushstring(pLuaState, "fe_context");
	lua_gettable(pLuaState, LUA_GLOBALSINDEX);
	LuaContext *pContext = (LuaContext *)(lua_touserdata(pLuaState,-1));

	sp<RecordGroup> spRG(new RecordGroup());
	t_lua_group::push(pContext, spRG);

	return 1;
}

int LuaContext::createRecord(lua_State *pLuaState)
{
	lua_pushstring(pLuaState, "fe_context");
	lua_gettable(pLuaState, LUA_GLOBALSINDEX);
	LuaContext *pContext = (LuaContext *)(lua_touserdata(pLuaState,-1));

	Record record;
	t_lua_record::push(pContext, record);

	return 1;
}

int LuaContext::createRecordArray(lua_State *pLuaState)
{
	lua_pushstring(pLuaState, "fe_context");
	lua_gettable(pLuaState, LUA_GLOBALSINDEX);
	LuaContext *pContext = (LuaContext *)(lua_touserdata(pLuaState,-1));

	sp<RecordArray> spRA(new RecordArray());
	t_lua_array::push(pContext, spRA);

	return 1;
}

int LuaContext::createScope(lua_State *pLuaState)
{
	lua_pushstring(pLuaState, "fe_context");
	lua_gettable(pLuaState, LUA_GLOBALSINDEX);
	LuaContext *pContext = (LuaContext *)(lua_touserdata(pLuaState,-1));

	sp<Scope> spScope(new Scope(*(pContext->master())));
	t_lua_scope::push(pContext, spScope);

	return 1;
}

int LuaContext::createScopeComponent(lua_State *pLuaState)
{
	t_lua_scope *pObj = (t_lua_scope *)lua_touserdata(pLuaState, -1);
	if(!pObj)
	{
		return 0;
	}
	lua_pop(pLuaState, 1);

	lua_pushstring(pLuaState, "fe_context");
	lua_gettable(pLuaState, LUA_GLOBALSINDEX);
//	LuaContext *pContext = (LuaContext *)(lua_touserdata(pLuaState,-1));

	sp<Component> spComponent = pObj->data();
	if(!spComponent.isValid())
	{
		lua_pushnil(pLuaState);
		return 1;
	}

	LuaComponentObject::pushComponent(pLuaState,spComponent);

	return 1;
}

void LuaContext::error(lua_State *pLuaState, const char *message)
{
	lua_pushstring(pLuaState, message);
	lua_error(pLuaState);
}

int LuaContext::debug(lua_State *pLuaState)
{
	String err = "\nLua Script Failed:\n[31m[1m";
	err.cat(lua_tostring(pLuaState, -1));
	err.cat("[0m\n");
	lua_pop(pLuaState, 1);

	int level = 0;
	lua_Debug ar;

	String message;
	while(lua_getstack(pLuaState, level, &ar))
	{
		lua_getinfo(pLuaState, "nSl", &ar);

		// name
		String name = "---";
		if(ar.name) { name = ar.name; }

		// namewhat
		String namewhat = "unknown";
		if(ar.namewhat) { namewhat = ar.namewhat; }

		// source
		String source = ar.source;

		// what
		String what = ar.what;

		if(ar.currentline >= 0)
		{
			message.sPrintf(
				"lua: [36m%-45s [0mline:[36m%5d [34m%-10s[0m",
				name.c_str(),
				ar.currentline,
				what.c_str()
				);
			err.cat(message);
			err.cat("\n");
		}

		U32 lineNumber=1;
		String buffer=source;
		while(!buffer.empty())
		{
			const BWORD highlight=(lineNumber==ar.currentline);
			const String line=buffer.parse("\"","\n",TRUE);

			message.sPrintf("%s%3d %s%s\n",
					highlight? "[33m": "",
					lineNumber,line.c_str(),
					highlight? "[0m": "");
			err.cat(message);
			lineNumber++;
		}
		err.cat("\n");

		level += 1;
	}

	lua_pushstring(pLuaState, err.c_str());
	return 1;
}

int LuaContext::manage(lua_State *pLuaState)
{
	String libname = lua_tostring(pLuaState, -1);
	lua_pop(pLuaState, 1);

	lua_pushstring(pLuaState, "fe_master");
	lua_gettable(pLuaState, LUA_GLOBALSINDEX);
	sp<Master> spMaster((Master *)(lua_touserdata(pLuaState,-1)));

	if(!successful(spMaster->registry()->manage(libname)))
	{
		lua_pushboolean(pLuaState, 0);
		return 1;
	}
	lua_pushboolean(pLuaState, 1);
	return 1;
}

int LuaContext::create(lua_State *pLuaState)
{
	String componentName = lua_tostring(pLuaState, -1);
	lua_pop(pLuaState, 1);
	if(componentName == "")
	{
		lua_pushnil(pLuaState);
		return 1;
	}

	lua_pushstring(pLuaState, "fe_context");
	lua_gettable(pLuaState, LUA_GLOBALSINDEX);
	LuaContext *pContext = (LuaContext *)(lua_touserdata(pLuaState,-1));

	sp<Component> spComponent =
		pContext->master()->registry()->create(componentName);
	if(!spComponent.isValid())
	{
		lua_pushnil(pLuaState);
		return 1;
	}

	LuaComponentObject::pushComponent(pLuaState,spComponent);

	return 1;
}

int LuaContext::catalog(lua_State *pLuaState)
{
	String interface;
	sp<Component> spC;
	if(lua_isstring(pLuaState, 1))
	{
		interface = lua_tostring(pLuaState, 1);
		if(interface == "")
		{
			lua_pushnil(pLuaState);
			return 1;
		}
	}
	else
	{
		spC = component<Component>(pLuaState, 1);
		if(!spC.isValid())
		{
			lua_pushnil(pLuaState);
			return 1;
		}
	}

	String label = lua_tostring(pLuaState, 2);
	if(label == "")
	{
		lua_pushnil(pLuaState);
		return 1;
	}

	lua_pushstring(pLuaState, "fe_context");
	lua_gettable(pLuaState, LUA_GLOBALSINDEX);
	LuaContext *pContext = (LuaContext *)(lua_touserdata(pLuaState,-1));

	sp<Component> spComponent;
	if(!spC.isValid())
	{
		spComponent = pContext->master()->catalog()->catalogComponent(
				interface, label);
	}
	else
	{
		spComponent = spC;
		pContext->master()->catalog()->catalogComponent(spC, label);
	}
	if(!spComponent.isValid())
	{
		lua_pushnil(pLuaState);
		return 1;
	}

	LuaComponentObject::pushComponent(pLuaState,spComponent);

	return 1;
}

int LuaContext::b_or(lua_State *pLuaState)
{
	FE_UWORD val = 0;
	int argc = lua_gettop(pLuaState);
	for(int i = 0; i < argc; i++)
	{
		val |= (int)lua_tonumber(pLuaState, i+1);
	}
	lua_pushnumber(pLuaState, val);
	return 1;
}

int LuaContext::b_and(lua_State *pLuaState)
{
	FE_UWORD val = ~0;
	int argc = lua_gettop(pLuaState);
	for(int i = 0; i < argc; i++)
	{
		val &= (int)lua_tonumber(pLuaState, i+1);
	}
	lua_pushnumber(pLuaState, val);
	return 1;
}

#define FEWEV FE_EVENT_LAYOUT

void LuaContext::setWindowEventEnums(void)
{
	set(FEWEV,"sourceNull",				WindowEvent::e_sourceNull);
	set(FEWEV,"sourceMousePosition",	WindowEvent::e_sourceMousePosition);
	set(FEWEV,"sourceMouseButton",		WindowEvent::e_sourceMouseButton);
	set(FEWEV,"sourceKeyboard",			WindowEvent::e_sourceKeyboard);
	set(FEWEV,"sourceSystem",			WindowEvent::e_sourceSystem);
	set(FEWEV,"sourceUser",				WindowEvent::e_sourceUser);
	set(FEWEV,"sourceAny",				WindowEvent::e_sourceAny);

	set(FEWEV,"itemNull",				WindowEvent::e_itemNull);
	set(FEWEV,"itemAny",				WindowEvent::e_itemAny);

	set(FEWEV,"itemMove",				WindowEvent::e_itemMove);
	set(FEWEV,"itemDrag",				WindowEvent::e_itemDrag);

	set(FEWEV,"itemLeft",				WindowEvent::e_itemLeft);
	set(FEWEV,"itemMiddle",				WindowEvent::e_itemMiddle);
	set(FEWEV,"itemRight",				WindowEvent::e_itemRight);
	set(FEWEV,"itemWheel",				WindowEvent::e_itemWheel);

	set(FEWEV,"keyBackspace",			WindowEvent::e_keyBackspace);
	set(FEWEV,"keyEscape",				WindowEvent::e_keyEscape);
	set(FEWEV,"keyDelete",				WindowEvent::e_keyDelete);
	set(FEWEV,"keyAsciiMask",			WindowEvent::e_keyAsciiMask);
	set(FEWEV,"keyCursor",				WindowEvent::e_keyCursor);
	set(FEWEV,"keyCursorUp",			WindowEvent::e_keyCursorUp);
	set(FEWEV,"keyCursorDown",			WindowEvent::e_keyCursorDown);
	set(FEWEV,"keyCursorLeft",			WindowEvent::e_keyCursorLeft);
	set(FEWEV,"keyCursorRight",			WindowEvent::e_keyCursorRight);
	set(FEWEV,"keyCursorMask",			WindowEvent::e_keyCursorMask);
	set(FEWEV,"keyShift",				WindowEvent::e_keyShift);
	set(FEWEV,"keyCapLock",				WindowEvent::e_keyCapLock);
	set(FEWEV,"keyControl",				WindowEvent::e_keyControl);
	set(FEWEV,"keyUsed",				WindowEvent::e_keyUsed);

	set(FEWEV,"itemIdle",				WindowEvent::e_itemIdle);
	set(FEWEV,"itemResize",				WindowEvent::e_itemResize);
	set(FEWEV,"itemExpose",				WindowEvent::e_itemExpose);
	set(FEWEV,"itemCloseWindow",		WindowEvent::e_itemCloseWindow);
	set(FEWEV,"itemQuit",				WindowEvent::e_itemQuit);

	set(FEWEV,"windowClear",			WindowEvent::e_windowClear);
	set(FEWEV,"windowSwapBuffers",		WindowEvent::e_windowSwapBuffers);

	set(FEWEV,"stateNull",				WindowEvent::e_stateNull);
	set(FEWEV,"statePress",				WindowEvent::e_statePress);
	set(FEWEV,"stateRelease",			WindowEvent::e_stateRelease);
	set(FEWEV,"stateRepeat",			WindowEvent::e_stateRepeat);
	set(FEWEV,"stateAny",				WindowEvent::e_stateAny);

	set(FEWEV,"mbNone",					WindowEvent::e_mbNone);
	set(FEWEV,"mbLeft",					WindowEvent::e_mbLeft);
	set(FEWEV,"mbMiddle",				WindowEvent::e_mbMiddle);
	set(FEWEV,"mbRight",				WindowEvent::e_mbRight);
	set(FEWEV,"mbAll",					WindowEvent::e_mbAll);
	set(FEWEV,"mbAny",					WindowEvent::e_mbAny);
}

} /* namespace */
